



pytensor-cache help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
python check-for-warnings.py allowed-warnings-base.txt
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest tests/test_breakpoint.py tests/test_config.py tests/test_gradient.py tests/test_ifelse.py tests/test_raise_op.py tests/test_rop.py tests/test_updates.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
